#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

# Manuals:
# PyOpenGL 3.0.2 Object Ref:
#   http://pyopengl.sourceforge.net/documentation/manual-3.0/index.xhtml
# OGL Programming Guide:
#   http://www.glprogramming.com/red/index.html
# GLUT Guide:
#   http://www.opengl.org/documentation/specs/glut/spec3/node1.html
#
# Specific Help:
#   Cyl between two pts:
#       http://www.thjsmith.com/40/cylinder-between-two-points-opengl-c
#   FSAA:
#       http://www.cs.uiuc.edu/class/sp05/cs419/accum/antialias.c
#       http://www.glprogramming.com/red/chapter06.html#name2
#
# Possible Extension Packages:
#   http://www.slicer.org
#   http://www.paraview.org/paraview/project/features.html

''' This module facilitates the manipulation of GL objects by maintaining a
list of GL calls to be supplied to a GL renderer.  These objects are pure
python and can therefore be easily serialized for passing between Nodes. '''


import copy
import traceback
import numpy as np  # for vector calcs
import numpy.linalg

from gpi import QtWidgets, QtGui, QtCore
from .logger import manager
# start logger for this module
log = manager.getLogger(__name__)

try:
    try:
        import OpenGL, OpenGL.GL, OpenGL.GLU, OpenGL.GLUT
        from OpenGL import GL, GLU
    
    # fallback for OpenGl Module import source
    except ImportError:
        # https://stackoverflow.com/questions/63475461/unable-to-import-opengl-gl-in-python-on-macos
        print('Patching OpenGL import for Big Sur (OSX)')
        from ctypes import util
        orig_util_find_library = util.find_library
        
        # retrieving the OpenGl Modules from its directory instead of cache
        def new_util_find_library (name):
            res = orig_util_find_library (name)
            if res: return res
            return '/System/Library/Frameworks/' + name + '.framework/' + name

        util.find_library = new_util_find_library
        
        import OpenGL, OpenGL.GL, OpenGL.GLU, OpenGL.GLUT
        from OpenGL import GL, GLU

except ImportError:
    log.warn('OpenGL was not found, GL objects and windows will not be supported in this session.')
    raise



class ObjectList(object):
    def __init__(self, oList=None):

        self._olist = []
        self._olist_noncacheable = []

        # keep track of clipping planes only one instance of each
        # can exist at a time so replace them.
        self._clipplanes = {}

        if oList is not None:
            self._olist = copy.deepcopy(oList.getCacheableList())
            self._olist_noncacheable = copy.deepcopy(oList.getNonCacheableList())
            self._clipplanes = copy.deepcopy(oList.getClipPlanes())

    def append(self, obj):

        # all special object hooks
        if isinstance(obj, ClipPlane):
            self._clipplanes[obj.getPlaneLabel()] = obj
            return

        if not obj.cacheable():
            self._olist_noncacheable.append(obj)
            return

        # all normal objects get appended
        self._olist.append(obj)

    def getCacheableList(self):
        return self._olist

    def getNonCacheableList(self):
        return self._olist_noncacheable

    def getClipPlanes(self):
        return self._clipplanes

    def len(self):
        return len(self._clipplanes) + len(self._olist) + len(self._olist_noncacheable)


class GPIGLObject(object):

    def __init__(self):
        # the name of THIS object (for referencing)
        self._label = self.__class__
        self._cacheable = True
        self._glwdg = None

        # material
        self._face = 'GL_FRONT'  # GL_FRONT_AND_BACK
        self._pname = 'GL_AMBIENT_AND_DIFFUSE'
        self._RGBA = (1.0, 1.0, 1.0, 1.0)  # default color is WHITE

        # shading
        self._shademodel = 'GL_SMOOTH'

        # subdivisions
        self._subdiv = 8

        # position (x, y, z)
        self._position = None

        # rotations individual matrices (x, y, z)
        self._rotxyz = None

        # rotations angle-vector (ang, x, y, z)
        self._rotav = None

        # scale (x, y, z)
        self._scale = None

        # multiple shape instances
        self._multiples = None

    def GL(self, attr):
        return getattr(GL, attr)

    def GLU(self, attr):
        return getattr(GLU, attr)

    def GLUT(self, attr):
        return getattr(GLUT, attr)

    def QtGui(self, attr):
        return getattr(QtGui, attr)

    def QtWidgets(self, attr):
        return getattr(QtWidgets, attr)

    def QtCor(self, attr):
        return getattr(QtCor, attr)

    def setGLWidgetRef(self, ref):
        self._glwdg = ref

    def cacheable(self):
        return self._cacheable

    def setMultiples(self, pos):
        '''Takes a list of positions specific to subclass shape.
        Either a list or a numpy array
        '''
        self._multiples = np.array(pos)

    def setSubdiv(self, sub):
        self._subdiv = sub

    def getSubdiv(self):
        return self._subdiv

    def setScale(self, scl):
        self._scale = scl

    def getScale(self):
        return self._scale

    def setPos(self, pos):
        self._position = pos

    def getPos(self):
        return self._position

    def setRotAV(self, rot):
        self._rotav = rot

    def getRotAV(self):
        return self._rotav

    def setRotXYZ(self, rot):
        self._rotxyz = rot

    def getRotXYZ(self):
        return self._rotxyz

    def setRGBA(self, RGBA):
        self._RGBA = RGBA

    def getRGBA(self):
        return self._RGBA

    def applyLighting(self):
        if self._RGBA:
            GL.glMaterialfv(self.GL(self._face), self.GL(self._pname), self._RGBA)
            GL.glShadeModel(self.GL(self._shademodel))

    def applyTransforms(self):
        GL.glPushMatrix()

        if self._position is not None:
            GL.glTranslatef(self._position[0], self._position[1], self._position[2])

        if self._rotav is not None:
            GL.glRotatef(self._rotav[0], self._rotav[1], self._rotav[2], self._rotav[3])

        if self._rotxyz is not None:  # independent axis rotation
            GL.glRotatef(self._rotxyz[0], 1.0, 0.0, 0.0)
            GL.glRotatef(self._rotxyz[1], 0.0, 1.0, 0.0)
            GL.glRotatef(self._rotxyz[2], 0.0, 0.0, 1.0)

        if self._scale is not None:
            GL.glScale(self._scale[0], self._scale[1], self._scale[2])

        try:
            self.applyShape()
        except:
            log.node(str(self.__class__)+':'+traceback.format_exc())

        GL.glPopMatrix()

    def applyShape(self):
        # subclass to implement specific quadrics etc...
        pass

    def run(self):
        if self._multiples is not None:
            self.applyLighting()
            for i in range(self._multiples.shape[0]):
                self._position = self._multiples[i].tolist()
                self.applyTransforms()
        else:
            self.applyLighting()
            self.applyTransforms()


class Text(GPIGLObject):

    def __init__(self):
        super(Text, self).__init__()

        self._cacheable = False
        self._text = 'Text'

        # Qt font
        self._font = 'Times New Roman'
        self._ptsize = 20

    def setFont(self, f):
        self._font = f

    def setSize(self, s):
        self._ptsize = s

    def setText(self, t):
        self._text = t

    def applyShape(self):
        '''Grab external QT rendering command.
        '''
        GL.glColor4d(*self._RGBA)
        f = QtGui.QFont(str(self._font), self._ptsize)
        p = self._position
        if self._glwdg is None:
            log.critical('Reference to GLWidget is not set! Aborting render.')
        else:
            self._glwdg.renderText(p[0], p[1], p[2], self._text, f)


class Sphere(GPIGLObject):

    def __init__(self):
        super(Sphere, self).__init__()

        # normals
        self._normals = 'GL_SMOOTH'
        self._texture = 'GL_TRUE'

        self._radius = 1.0

    def setNormals(self, n):
        self._normals = n

    def setTexture(self, t):
        self._texture = t

    def setRadius(self, r):
        self._radius = r

    def applyShape(self):
        Q=GLU.gluNewQuadric()
        GLU.gluQuadricNormals(Q, self.GL(self._normals))
        GLU.gluQuadricTexture(Q, self.GL(self._texture))
        GLU.gluSphere(Q, self._radius, self._subdiv, self._subdiv/2)
        GLU.gluDeleteQuadric( Q )

class Cylinder(GPIGLObject):

    def __init__(self):
        super(Cylinder, self).__init__()

        # normals
        self._normals = 'GL_SMOOTH'
        self._texture = 'GL_TRUE'

        self._base = 1.0
        self._top = 1.0
        self._height = 1.0
        self._z = np.array([0,0,1.0])
        self._conv = 180.0 / np.pi

        self._endToEnd = False

        self._Quadric = None
        self._f1 = None
        self._f2 = None
        self._f3 = None

    def setEndToEnd(self, val):
        self._endToEnd = val

    def setP1P2(self, p1, p2):
        p1 = np.array(p1)
        self._position = np.array(p2)

        p = p1 - self._position
        t = np.cross(self._z, p)
        self._height = np.linalg.norm(p)

        ang = self._conv * np.arccos(p[2]/self._height)
        self._rotav = (ang, t[0], t[1], t[2])

    def setBase(self, base):
        self._base = base

    def getBase(self):
        return self._base

    def setTop(self, top):
        self._top = top

    def getTop(self):
        return self._top

    def setHeight(self, h):
        self._height = h

    def getHeight(self):
        return self._height

    def setNormals(self, n):
        self._normals = n

    def setTexture(self, t):
        self._texture = t

    def setRadius(self, r):
        self._top = r
        self._base = r

    def applyShape(self):
        self._Quadric=GLU.gluNewQuadric()
        GLU.gluQuadricNormals(self._Quadric, self.GL(self._normals))
        GLU.gluQuadricTexture(self._Quadric, self.GL(self._texture))
        GLU.gluCylinder(self._Quadric, self._base, self._top, self._height, self._subdiv, 1)
        GLU.gluDeleteQuadric(self._Quadric)
        self._Quadric = None

    def cleanupCache(self):
        GLU.gluDeleteQuadric(self._Quadric)
        self._Quadric = None

    def run(self):
        # TODO: this can be made alot faster by placing these Quadric calls here, however, this needs to be compatible with the Axes object.
        # -make a better baseclass for these cases.

        #self._Quadric=GLU.gluNewQuadric()
        #self._f1 = lambda: GLU.gluQuadricNormals(self._Quadric, self.GL(self._normals))
        #self._f2 = lambda: GLU.gluQuadricTexture(self._Quadric, self.GL(self._texture))


        self.applyLighting()
        #self._f1()
        #self._f2()

        if self._multiples is not None:
            if self._endToEnd:
                for i in range(self._multiples.shape[0]-1):
                    self.setP1P2(self._multiples[i], self._multiples[i+1])
                    self.applyTransforms()
            else:
                for pos in self._multiples:
                    self._position = pos
                    self.applyTransforms()
        else:
            self.applyTransforms()

        # remove quadrics and such
        #self.cleanupCache()


class ClipPlane(GPIGLObject):

    def __init__(self):
        super(ClipPlane, self).__init__()

        # six planes max
        self._planeNo = 0

        # normal
        self._norm = 'x'

    def setNorm(self, n):
        self._norm = n

    def setPlaneNum(self, no):
        if (no < 6) and (no >= 0):
            self._planeNo = int(no)

    def getPlaneLabel(self):
        return 'GL_CLIP_PLANE'+str(self._planeNo)

    def applyShape(self):
        p = self.GL(self.getPlaneLabel())
        GL.glEnable(p)
        if self._norm == 'x':
            GL.glClipPlane(p, (1,0,0,0))
        if self._norm == 'y':
            GL.glClipPlane(p, (0,1,0,0))
        if self._norm == 'z':
            GL.glClipPlane(p, (0,0,1,0))

class Axes(GPIGLObject):

    def __init__(self):
        super(Axes, self).__init__()

        self._RGBA = None
        self._cacheable = False

        self._x = Cylinder()
        self._y = Cylinder()
        self._z = Cylinder()

        self._xl = Text()
        self._yl = Text()
        self._zl = Text()

        self._x.setRGBA((1.0, 0.0, 0.0, 1.0))  # RED
        self._y.setRGBA((0.0, 1.0, 0.0, 1.0))  # GREEN
        self._z.setRGBA((0.0, 0.0, 1.0, 1.0))  # BLUE

        self._xl.setRGBA((1.0, 0.0, 0.0, 1.0))  # RED
        self._yl.setRGBA((0.0, 1.0, 0.0, 1.0))  # GREEN
        self._zl.setRGBA((0.0, 0.0, 1.0, 1.0))  # BLUE

        self._x.setP1P2((-1,0,0), (1,0,0))
        self._y.setP1P2((0,-1,0), (0,1,0))
        self._z.setP1P2((0,0,1), (0,0,-1))  # NRZ flipped for Linux

        self._a = 0.1
        self._xl.setPos((self._x.getHeight()/4 +self._a, 0, 0))
        self._yl.setPos((0, self._y.getHeight()/4 +self._a, 0))
        self._zl.setPos((0, 0, self._z.getHeight()/4 +self._a))

        self._xl.setText('x')
        self._yl.setText('y')
        self._zl.setText('z')

    def setGLWidgetRef(self, ref):
        self._xl.setGLWidgetRef(ref)
        self._yl.setGLWidgetRef(ref)
        self._zl.setGLWidgetRef(ref)

    def setRGBA(self, c):
        pass

    def setSubdiv(self, s):
        self._x.setSubdiv(s)
        self._y.setSubdiv(s)
        self._z.setSubdiv(s)

    def setRadius(self, r):
        self._x.setP1P2((-r,0,0), (r,0,0))
        self._y.setP1P2((0,-r,0), (0,r,0))
        self._z.setP1P2((0,0,r), (0,0,-r))  # NRZ flipped for Linux

        self._xl.setPos((self._x.getHeight()/4 +self._a, 0, 0))
        self._yl.setPos((0, self._y.getHeight()/4 +self._a, 0))
        self._zl.setPos((0, 0, self._z.getHeight()/4 +self._a))

    def setTubeRadius(self, r):
        self._x.setTop(r)
        self._x.setBase(r)
        self._y.setTop(r)
        self._y.setBase(r)
        self._z.setTop(r)
        self._z.setBase(r)

        self._xl.setSize(200.0*r)
        self._yl.setSize(200.0*r)
        self._zl.setSize(200.0*r)

    def applyShape(self):
        self._x.applyLighting()
        self._x.applyTransforms()
        self._y.applyLighting()
        self._y.applyTransforms()
        self._z.applyLighting()
        self._z.applyTransforms()

        self._xl.applyLighting()
        self._xl.applyTransforms()
        self._yl.applyLighting()
        self._yl.applyTransforms()
        self._zl.applyLighting()
        self._zl.applyTransforms()




#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.


from gpi import QtCore, QtGui, QtWidgets

# gpi
from .port import InPort, OutPort, Port
from .edge import Edge
from .macroNode import PortEdge
from .node import Node
from .defines import GPI_PORT_EVENT
from .defines import getKeyboardModifiers, printMouseEvent
from .logger import manager

# start logger for this module
log = manager.getLogger(__name__)


class CanvasScene(QtWidgets.QGraphicsScene):
    '''Supports the main canvas widget by drawing shapes for displaying
    interactions between elements on the canvas (e.g. selecting nodes). '''

    def __init__(self, parent=None):
        super(CanvasScene, self).__init__(parent)

        self.graph = parent
        self.line = None
        self.rubberBand = None
        self.origin = QtCore.QPointF()

        # during a port connection this is used to hold
        # a list of matching port types for highlighting
        self.portMatches = []

    def startLineDraw(self, event):
        # highlight all ports that match
        startItems = self.items(event.scenePos())
        if len(startItems):
            if isinstance(startItems[0], OutPort):
                if QtCore.Qt.AltModifier == getKeyboardModifiers():
                    startItems[0].toggleMemSaver()
                    return
                self.portMatches = startItems[0].findMatchingInPorts()
                for port in self.portMatches:
                    port.scaleUp()
                self.graph.viewAndSceneForcedUpdate()

            if isinstance(startItems[0], InPort):
                self.portMatches = startItems[0].findMatchingOutPorts()
                for port in self.portMatches:
                    port.scaleUp()
                self.graph.viewAndSceneForcedUpdate()

        self.line = QtWidgets.QGraphicsLineItem(
            QtCore.QLineF(event.scenePos(), event.scenePos()))
        fade = QtGui.QColor(QtCore.Qt.red)
        fade.setAlpha(150)
        self.line.setPen(QtGui.QPen(fade, 2))
        self.line.setZValue(10)
        self.addItem(self.line)

    def midLineDraw(self, event):

        newLine = QtCore.QLineF(self.line.line().p1(), event.scenePos())
        self.line.setLine(newLine)

        # attach source or sink connection
        #if QtCore.Qt.AltModifier == getKeyboardModifiers():
        #    self.line.setPen(QtGui.QPen(QtCore.Qt.blue, 2))
        #    self.dumpCursorStack()
        #    QtWidgets.QApplication.setOverrideCursor(
        #        QtGui.QCursor(QtCore.Qt.CrossCursor))
        #    return
        #else:
        #    self.line.setPen(QtGui.QPen(QtCore.Qt.red, 2))

        startItems = self.items(event.scenePos())
        if len(startItems) and startItems[0] == self.line:
            startItems.pop(0)

        self.dumpCursorStack()
        QtWidgets.QApplication.setOverrideCursor(
            QtGui.QCursor(QtCore.Qt.ForbiddenCursor))

        if len(startItems):
            if isinstance(startItems[0], Port):
                self.dumpCursorStack()
                QtWidgets.QApplication.setOverrideCursor(
                    QtGui.QCursor(QtCore.Qt.CrossCursor))

    def endLineDraw(self, event):
        self.dumpCursorStack()
        QtWidgets.QApplication.setOverrideCursor(
            QtGui.QCursor(QtCore.Qt.OpenHandCursor))
        QtWidgets.QApplication.restoreOverrideCursor()

        startItems = self.items(self.line.line().p1())

        # if there are no valid start items then mark for removal
        if len(startItems) and startItems[0] == self.line:
            startItems.pop(0)

        # reset any hovering events
        if len(startItems):
            if isinstance(startItems[0], Port):
                startItems[0].hoverLeaveEvent(event)

        # if there are no valid end items then mark for removal
        endItems = self.items(self.line.line().p2())
        if len(endItems) and endItems[0] == self.line:
            endItems.pop(0)

        # remove temporary drag line
        self.removeItem(self.line)
        self.line = None

        # if there are valid start and end items then continue
        if len(startItems) and len(endItems):

            # connect ports
            if isinstance(startItems[0], Port) and isinstance(endItems[0], Port) \
                    and startItems[0] != endItems[0] and type(startItems[0]) != type(endItems[0]):

                # find which is which
                if isinstance(startItems[0], InPort):
                    inport = startItems[0]
                    outport = endItems[0]
                else:
                    inport = endItems[0]
                    outport = startItems[0]

                # check if the ports belong to the same node
                if inport.node == outport.node:
                    log.warn("CanvasScene: inport and outport belong to the same node.")
                    super(CanvasScene, self).mouseReleaseEvent(event)
                    return

                # check the number of connections on the inport
                # if one exists then skip
                if len(inport.edges()) > 0:
                    log.warn("CanvasScene: inport occupied, connection dropped")
                    super(CanvasScene, self).mouseReleaseEvent(event)
                    return

                # This, currently, is the only way edges know
                # which port is actually a source or dest.
                # NOTE: Network description files will have to enforce this.
                newEdge = Edge(outport, inport)
                self.addItem(newEdge)

                # if its cyclic then don't allow the connection
                # -at the same time, calculate the hierarchy
                nodeHierarchy = inport.getNode().graph.calcNodeHierarchy()
                if nodeHierarchy is None:
                    self.removeItem(newEdge)
                    newEdge.detachSelf()
                    # del newEdge
                    log.warn("CanvasScene: cyclic, connection dropped")
                else:
                    # CONNECTION ADDED
                    # Since node hierarchy is recalculated, also
                    # take the time to flag nodes for processing
                    # 1) check for matching spec type
                    if not (inport.checkUpstreamPortType()):
                        self.removeItem(newEdge)
                        newEdge.detachSelf(update=False)
                        # del newEdge
                        log.warn("CanvasScene: data type mismatch, connection dropped")
                    else:
                        # 2) set the downstream node's pending_event
                        inport.getNode().setEventStatus({GPI_PORT_EVENT: inport.portTitle})

                        # trigger a force recalculation
                        inport.getNode().graph.itemMoved()

                        # trigger name update
                        inport.getNode().refreshName()
                        outport.getNode().refreshName()

                        # trigger event queue, if its idle
                        inport.getNode().graph._switchSig.emit('check')

                        if len(self.portMatches):
                            for port in self.portMatches:
                                port.resetScale()
                            self.portMatches = []

                        inport.edges()[0].adjust()
                        for edge in outport.edges():
                            edge.adjust()

            # remove edge by drawing a line across it
            elif isinstance(startItems[0], Edge) and isinstance(startItems[0], Edge) and startItems[0] == endItems[0]:
                # remove from scene
                self.removeItem(startItems[0])
                # remove from ports
                startItems[0].detachSelf()
                # remove from memory
                # del startItems[0]
        # reset the port size changes during port connect.
        if len(self.portMatches):
            for port in self.portMatches:
                port.resetScale()
            self.portMatches = []
        self.line = None

    def dumpCursorStack(self):
        while QtWidgets.QApplication.overrideCursor():
            QtWidgets.QApplication.restoreOverrideCursor()

    def mousePressEvent(self, event):  # CANVAS SCENE
        printMouseEvent(self, event)
        modifiers = getKeyboardModifiers()

        # allow graphics view panning
        if self.graph._panning:
            event.ignore()
            return

        # if its not a port, then don't draw a line
        modmidbutton_event = (event.button() == QtCore.Qt.LeftButton
                              and modifiers == QtCore.Qt.AltModifier)

        if ((event.button() == QtCore.Qt.LeftButton) or (event.button() == QtCore.Qt.MidButton) or modmidbutton_event) \
                and isinstance(self.itemAt(event.scenePos(), QtGui.QTransform()), Port):
            event.accept()
            self.startLineDraw(event)
        # rubber band select
        # elif ((event.button() == QtCore.Qt.MidButton) or modmidbutton_event):
        elif ((event.button() == QtCore.Qt.LeftButton) \
                and not isinstance(self.itemAt(event.scenePos(), QtGui.QTransform()), Node) \
                and not isinstance(self.itemAt(event.scenePos(), QtGui.QTransform()), PortEdge)):
            event.accept()
            self.unselectAllItems()  # reset select before making another
            self.origin = event.scenePos()
            self.rubberBand = QtWidgets.QGraphicsRectItem(
                QtCore.QRectF(self.origin, QtCore.QSizeF()))
            self.rubberBand.setPen(QtGui.QPen(
                QtCore.Qt.gray, 0, QtCore.Qt.SolidLine))
            self.rubberBand.setBrush(QtGui.QBrush(QtCore.Qt.lightGray))
            self.rubberBand.setZValue(0)
            self.addItem(self.rubberBand)
        else:
            QtWidgets.QApplication.restoreOverrideCursor()
            event.ignore()
            super(CanvasScene, self).mousePressEvent(event)

    def mouseMoveEvent(self, event):  # CANVAS SCENE

        # allow graphics view panning
        if self.graph._panning:
            event.ignore()
            return

        if self.line:
            event.accept()
            self.midLineDraw(event)
        elif self.rubberBand:
            event.accept()
            newRect = QtCore.QRectF(self.origin, event.scenePos())
            self.rubberBand.setRect(newRect)
        else:
            event.ignore()
            self.dumpCursorStack()
            super(CanvasScene, self).mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):  # CANVAS SCENE

        # allow graphics view panning
        if self.graph._panning:
            event.ignore()
            return

        printMouseEvent(self, event)
        if self.line:
            event.accept()
            self.endLineDraw(event)
        elif self.rubberBand:
            event.accept()
            selarea = self.rubberBand.rect()
            self.setSelectedNodes(selarea)
            self.removeItem(self.rubberBand)
            self.rubberBand = None
        else:
            event.ignore()
            self.line = None
            super(CanvasScene, self).mouseReleaseEvent(event)

    def unselectAllItems(self):
        # TODO: add Z level changes here too
        for item in list(self.items()):
            if item.isSelected():
                item.setSelected(False)

    def setSelectedItems(self, box):
        self.unselectAllItems()
        for item in self.items(box):
            item.setSelected(True)

    def setSelectedNodes(self, box):
        self.unselectAllItems()
        for item in self.items(box):
            if isinstance(item, Node):
                item.setSelected(True)

    def makeOnlyTheseNodesSelected(self, nodes):
        # need to distinguish between nodes and other items
        self.unselectAllItems()
        for node in nodes:
            node.setSelected(True)
            node.update()



#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.


import gpi
from gpi import QtCore
from .logger import manager

# start logger for this module
log = manager.getLogger(__name__)


class GPIState(QtCore.QObject):
    """A single FSM state to be used with the GPI_FSM class.  It manages its
    transitions, entry and exit functions.
    """
    entered = gpi.Signal()
    exited = gpi.Signal()

    def __init__(self, name, func, machine=None, efunc=None):
        super(GPIState, self).__init__()
        if not isinstance(name, str):
            msg = "expecting str arg: GPIState.__init__(>str<,func,GPI_FSM)"
            log.critical(msg)
        if machine:
            if not isinstance(machine, GPI_FSM):
                msg = "expecting GPI_FSM arg: GPIState.__init__(str,func,>GPI_FSM<)"
                log.critical(msg)
        self._name = str(name)
        self._transitions = {}
        self._func = func  # on entry
        self._efunc = efunc  # on exit
        if machine:
            machine.addState(self)

    def addTransition(self, sig, state):
        if not isinstance(sig, str):
            msg = "expecting str arg: GPIState.addTransition(>str<,GPIState)"
            log.critical(msg)
        self._transitions[sig] = state

    @property
    def name(self):
        return self._name

    def transitions(self):
        return self._transitions

    def onEntry(self, sig):
        self.entered.emit()
        if self._func:
            self._func(sig)

    def onExit(self, sig):
        self.exited.emit()
        if self._efunc:
            self._efunc(sig)


class GPI_FSM(QtCore.QObject):
    """The GPI Canvase and Nodes operate within different states.  The finite
    state machine allows for fast and easy checking to determine whether the
    user has performed an invalid operation or not.
    """
    switched = gpi.Signal()

    def __init__(self, name=''):
        super(GPI_FSM, self).__init__()
        self._name = name
        self._states = []
        self._cur_state = None

    @property
    def curState(self):
        return self._cur_state

    @property
    def curStateName(self):
        return self._cur_state.name

    @property
    def name(self):
        return self._name

    def next(self, dsig):
        '''Validate input signal as part of _cur_state,
        then switch to indicated state.'''
        if isinstance(dsig, str):
            sig = str(dsig)
            dsig = sig
        elif isinstance(dsig, str):
            sig = str(dsig)
            dsig = sig
        elif isinstance(dsig, dict):
            if 'sig' in dsig:
                if isinstance(dsig['sig'], str) or isinstance(dsig['sig'], str):
                    sig = str(dsig['sig'])
                    dsig['sig'] = sig
                else:
                    msg = "expecting str in dict[\'sig\'] in arg: GPI_FSM(" + self._name + ").next(>str<)"
                    log.critical(msg)
            else:
                msg = "expecting str key \'sig\' in arg: GPI_FSM(" + self._name + ").next(>str<)"
                log.critical(msg)
                return
        else:
            msg = "expecting str (or str in dict[\'sig\']) arg: GPI_FSM(" + \
                    self._name + ").next(>str<) type:" + str(type(dsig))
            log.critical(msg)
            return

        if sig in self._cur_state.transitions():
            self._cur_state.onExit(dsig)
            self._cur_state = self._cur_state.transitions()[sig]
            msg = "GPI_FSM(" + self._name + "):next(): Switched to state(" + \
                self._cur_state.name + ")"
            log.debug(msg)
            self._cur_state.onEntry(dsig)
            self.switched.emit()
        else:
            msg = "GPI_FSM(" + self._name + "):next(): state(" + self._cur_state.name + \
                ") has no transition: \'" + str(
                    sig) + "\'\n" + str(self._cur_state.transitions())
            log.debug(msg)

    def addState(self, state):
        if state in self._states:
            msg = "EMPHATIC WARNING!!!: GPI_FSM(" + self._name + "):addState(): Warning, state(" + state.name \
                + ") already exists, skipping..."
            log.critical(msg)
        else:
            self._states.append(state)

    def start(self, state):
        if state in self._states:
            self._cur_state = state
            msg = "GPI_FSM(" + self._name + \
                "):start(): in state(" + state.name + ")"
            log.debug(msg)
            self._cur_state.onEntry('init')
        else:
            msg = "GPI_FSM(" + self._name + "):start(): ERROR, state(" + state.name \
                + ") state not in list, can't start GPI_FSM!"
            log.critical(msg)



#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.


from .defines import ExternalType
from .logger import manager
# start logger for this module
log = manager.getLogger(__name__)

GPITYPE_PASS = 'PASS'

def osuper(cls, obj):
    '''for some reason (possibly old-style-class), super() fails in this
    module when using multiple canvases.  So instead of calling super()
    just return a handle to a new GPIDefaultType() instance.
    '''
    try:
        return super(cls, obj)
    except:
        return GPIDefaultType()
    return self

class GPIDefaultType(object):
    """This default class is the base class of all GPITypes.  It
    provides default behavior for edgeTips and toolTips.  The
    default behavior for port and data type matching is to pass.
    """
    GPIType = ExternalType  # ensures the subclass is of THIS class

    def __init__(self):
        # setup logger for subclasses
        self.log = manager.getLogger(__name__)

    def edgeTip(self, data):
        """Returns brief information for posting
        on the connection pipe. A one liner.
        """
        return ""

    def toolTip_Data(self, data):
        """Returns brief information for posting
        in the port's tool tip. This has specifics
        about the data being held at the outport.
        """
        return str(type(data))

    def toolTip_Port(self):
        """Returns brief information for posting
        in the port's tool tip about the port
        enforcement.
        """
        return str(GPITYPE_PASS)

    def matchesType(self, type_cls):
        """Returns True if self matches type_cls.
        By default all types match.
        Fail-usable in case the porttype
        can't be found.  type_cls is of the
        upstream port since the inPort is the
        limiting factor.
        """
        return True

    def isFreeType(self, type_cls=None):
        """Determine whether the compared type
        is implementing pass behavior (which
        is currently the default).
        """

        # With no args, 'self' is checked.
        if type_cls is None:
            if type(self) == GPIDefaultType:
                return True
            return False

        # test input class
        if type(type_cls) == GPIDefaultType:
            return True
        return False

    def matchesData(self, indata):
        """Returns True if self matches input data.
        By default all data matches.
        Fail-usable in case the porttype
        can't be found.
        """
        return True

    def setDataAttr(self, data):
        """Set any attributes on the data object
        (e.g. numpy arrays need to be readonly).
        the data object can be modified and
        returned in this hook.
        Passthrough by default.
        """
        return data

    def setTypeParms(self, **kwargs):
        """Use the kwargs dict to set user
        specified args to add<In/Out>Port()
        """
        for k, v in list(kwargs.items()):
            setter = "set_"+k
            if hasattr(self, setter):
                getattr(self, setter)(v)
            else:
                log.warn('No attribute \''+str(k)+'\' in port-type \''+str(GPITYPE_PASS)+'\',\n\t\tCheck your requested port-type in addInPort() or addOutPort().')



#!/usr/bin/env python

#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

# Brief: a make script that can double as a setup script.

'''
Use python distutils to build extension modules.  This script can be called
directly from the commandline to build C-extensions or check pure python
extensions.

A C/C++ extension module that implements an alorithm or method.

    To make, issue the following command:
        $ ./make.py <basename>
        or
        $ ./make.py <basename>.cpp
        or
        $ ./make.py <basename>.py
'''
import subprocess
from setuptools import setup, Extension
import os
import sys
import optparse  # get and process user input args
import platform
import py_compile
import traceback
import numpy

from gpi.config import Config

# error codes
SUCCESS = 0
ERROR_FAILED_COMPILATION = 1
ERROR_NO_VALID_TARGETS = 2
ERROR_INVALID_RECURSION_DEPTH = 3
ERROR_LIBRARY_CONFLICT = 4
ERROR_EXTERNAL_APP = 5

print("\n"+str(sys.version)+"\n")

# from:
# http://stackoverflow.com/questions/287871/print-in-terminal-with-colors-using-python
class Cl:
    HDR = '\033[95m'
    OKBL = '\033[94m'
    OKGR = '\033[92m'
    WRN = '\033[93m'
    FAIL = '\033[91m'
    ESC = '\033[0m'

# The basic distutils setup().
def compile(mod_name, include_dirs=[], libraries=[], library_dirs=[],
            extra_compile_args=[], runtime_library_dirs=[]):

    print(("Making target: " + mod_name))

    # do usual generic module setup
    # NOT: 'mod_name' must have an init<name>
    # function defined in the .cpp code.
    Module1 = Extension(mod_name,
                        # define_macros = [('MAJOR_VERSION',
                        # '1'),('MINOR_VERSION', '0')],
                        include_dirs=list(set(include_dirs)),
                        libraries=list(set(libraries)),
                        library_dirs=list(set(library_dirs)),
                        extra_compile_args=list(set(extra_compile_args)),
                        runtime_library_dirs=list(set(runtime_library_dirs)),
                        sources=[mod_name + '_PyMOD.cpp'])

    # run the setup() function
    try:
        setup(name=mod_name,
              version='0.1-dev',
              description='A kcii library of algorithms and methods.',
              ext_modules=[Module1],
              script_args=["build_ext", "--inplace", "--force"])
    except:
        print((sys.exc_info()))
        print(("FAILED: " + mod_name))
        return 1

    print(("SUCCESS: " + mod_name))
    return 0


def packageArgs(args):
    """Split path and filename info into a dictionary.
    """
    cwd = os.getcwd()
    targets = []
    for arg in args:
        fn = os.path.splitext(os.path.basename(arg))[0]
        ext = os.path.splitext(os.path.basename(arg))[1]
        dn = os.path.dirname(arg)
        targets.append({'pth': cwd + '/' + dn, 'fn': fn, 'ext': ext})
    return targets

def isPythonPackageDir(path):
    return os.path.isfile(str(path)+'/__init__.py')

def findLibraries(basepath):
    # TODO: this searching should be combined with the search in library.py and
    # unified in config.py since they both need to know which libraries are
    # present.

    # if the basepath IS the library directory
    if isPythonPackageDir(basepath):
        return [basepath]

    # check for subdirectories
    libs = []
    for p in os.listdir(basepath):
        subdir = os.path.join(basepath,p)
        if os.path.isdir(subdir):
            if isPythonPackageDir(subdir):
                libs.append(subdir)
    return libs

def targetWalk(recursion_depth=1):
    """Recurse into directories and look for .cpp files to compile.
    TODO: check if the file is a valid python module.
    """
    targets = []
    ipath = os.getcwd()
    ocnt = ipath.count('/')
    for path, dn, fn in os.walk(ipath):
        if path.count('/') - ocnt <= recursion_depth:
            if len(fn):
                for fil in fn:

                    # only attempt _PyMOD.cpp
                    if fil.endswith(".cpp"):
                        if fil.endswith("_PyMOD.cpp"):
                            fn = os.path.splitext(fil)[0]
                            ext = os.path.splitext(fil)[1]
                            targets.append({'pth': path, 'fn': fn, 'ext': ext})

                    # byte-compile all .py files
                    if fil.endswith(".py"):
                        fn = os.path.splitext(fil)[0]
                        ext = os.path.splitext(fil)[1]
                        targets.append({'pth': path, 'fn': fn, 'ext': ext})

    return targets


def makePy(basename, ext, fmt=False, check_fmt=None):

    if check_fmt is None:
        check_fmt = []

    target = [basename, ext]

    # AUTOPEP8
    if fmt:
        try:
            import autopep8
            print(("\nFound: autopep8 " + str(autopep8.__version__) + "..."))
            print(("Reformatting Python script: " + "".join(target)))
            os.system('autopep8 -i --max-line-length 256 ' + "".join(target))
        except:
            print("Failed to perform auto-formatting \
                with \'autopep8\'. Do you have it installed?")

    if 'pep8' in check_fmt:
        # PEP8
        try:
            import pep8
            print(("\nFound: pep8 " + str(pep8.__version__) + "..."))
            print(("Checking Python script: " + "".join(target)))
            print(("pep8 found these problems with your code, START" + Cl.WRN))
            os.system('pep8 --count --statistics --show-source '
                      + "".join(target))
            print((Cl.ESC + "pep8 END"))
        except:
            print("Failed to perform check with \'pep8\'. Do you have it installed?")

    if 'pyflakes' in check_fmt:
        # PYFLAKES
        try:
            import pyflakes
            print(("\nFound: pyflakes " + str(pyflakes.__version__) + "..."))
            print(("Checking Python script: " + "".join(target)))
            print(("pyflakes found these problems with your code, START" + Cl.FAIL))
            os.system('pyflakes ' + "".join(target))
            print((Cl.ESC + "pyflakes END"))
        except:
            print("Failed to perform check with \'pyflakes\'. Do you have it installed?")

    # FORCE COMPILE
    try:
        print('\nAttemping py_compile...')
        py_compile.compile(''.join(target), doraise=True)
        print('py_compile END')
        print(('\nSUCCESS: '+''.join(target)))
        return 0
    except:
        print((Cl.FAIL + str(traceback.format_exc()) + Cl.ESC))
        print('py_compile END')
        print(('\nFAILED: '+''.join(target)))
        return 1


def make(GPI_PREFIX=None):
    '''Commandline interface to the make utilities.
    '''

    CWD = os.path.realpath('.')

    # LIBRARIES, INCLUDES, ENV-VARS
    include_dirs = []
    libraries = []
    library_dirs = []
    extra_compile_args = []  # ['--version']
    runtime_library_dirs = []

    import pathlib
    GPI_DIR = pathlib.Path(__file__).parent.resolve()
    print("Adding GPI include directory")
    if GPI_PREFIX is not None:
        include_dirs.append(os.path.join(GPI_PREFIX, 'include', 'eigen3'))
        include_dirs.append(os.path.join(GPI_PREFIX, 'include'))
        include_dirs.append(os.path.join(GPI_DIR, 'include'))
        if platform.system() == 'Windows':
            include_dirs.append(os.path.join(GPI_PREFIX, 'Library/include'))

    parser = optparse.OptionParser()
    parser.add_option('--preprocess', dest='preprocess', default=False,
                      action="store_true", help='''Only do preprocessing to \
                              target (the resulting .o file will be \
                              preprocessed code.)''')
    parser.add_option('-w', '--suppressWarnings', dest='suppressWarnings',
                      default=False, action="store_true",
                      help='''Tell gcc to only display errors.''')
    parser.add_option('--fmt', dest='format', default=False,
                      action="store_true",
                      help="Auto-format using the autopep8 and astyle scripts.")
    parser.add_option('--pep8', dest='check_format',
                      action="append_const", const="pep8",
                      help="Check Python code format using pep8.")
    parser.add_option('--pyflakes', dest='check_format',
                      action="append_const", const="pyflakes",
                      help="Check Python code format using pyflakes.")
    parser.add_option('--all', dest='makeall', default=False,
                      action='store_true',
                      help="Recursively search for .cpp files and attempt to" +
                      "make them (integer arg sets recursion depth).")
    parser.add_option('-r', '--rdepth', dest='makeall_rdepth', type="int",
                      default=1,
                      help="Integer arg sets recursion depth for makeall.")
    parser.add_option('--debug', dest='debug', default=False,
                      action='store_true',
                      help="Uses range checker for PyFI::Array calls.")
    parser.add_option('--ignore-gpirc', dest='ignore_gpirc', default=False,
                      action='store_true',
                      help="Ignore the ~/.gpirc config.")
    parser.add_option('--ignore-system-libs', dest='ignore_sys', default=False,
                      action='store_true',
                      help="Ignore the system libraries (e.g. for conda build).")
    parser.add_option('--osx-ver', dest='osx_target_ver',
                      help="Override tgt. version for OSX builds (must be '10.X').")

    parser.add_option(
        '-v', '--verbose', dest='verbose', default=False, action="store_true",
        help='''Verbosity.''')

    parser.add_option(
        '-d', '--distdebug', dest='distdebug', default=False, action="store_true",
        help='''Sets DISTUTILS_DEBUG. ''')

    # get user input 'options', and extra 'args' that were unprocessed
    options, args = parser.parse_args()
    opt = vars(options)

    # debug the distutils setup()
    if options.distdebug:
        os.environ['DISTUTILS_DEBUG'] = '1'

    # gather args and their path info
    targets = None
    if len(args):
        targets = packageArgs(args)

    # gather all _PyMOD.cpp and .py files
    # supersedes 'args' if present.
    if options.makeall:
        if options.makeall_rdepth < 0:
            print((Cl.FAIL + "ERROR: recursion depth is set to an invalid number." + Cl.ESC))
            sys.exit(ERROR_INVALID_RECURSION_DEPTH)
        targets = targetWalk(options.makeall_rdepth)

    if targets is None:
        print((Cl.FAIL + "ERROR: no targets specified." + Cl.ESC))
        sys.exit(ERROR_NO_VALID_TARGETS)

    default_cpp = True
    if options.ignore_gpirc:
        print('Ignoring the ~/.gpirc...')
    else:
    # USER MAKE config
        if (len(Config.MAKE_CFLAGS) + len(Config.MAKE_LIBS) + len(Config.MAKE_INC_DIRS) + len(Config.MAKE_LIB_DIRS)) > 0:
            print("Adding USER include dirs")
            # add user libs
            libraries += Config.MAKE_LIBS
            include_dirs += Config.MAKE_INC_DIRS
            library_dirs += Config.MAKE_LIB_DIRS
            extra_compile_args += Config.MAKE_CFLAGS
        if any("c++" in cflag for cflag in Config.MAKE_CFLAGS):
            default_cpp = False

    # Anaconda environment includes
    # includes FFTW and eigen
    print("Adding Anaconda lib and inc dirs...")
    if platform.system() == 'Windows':
        library_dirs += [os.path.join(GPI_PREFIX, 'Library/lib')]
    else:
        library_dirs += [os.path.join(GPI_PREFIX, 'lib')]
    include_dirs += [numpy.get_include()]

    # GPI library dirs
    print("Adding GPI library dirs")
    # add libs from library paths
    found_libs = {}
    search_dirs = []
    if not options.ignore_gpirc:
        print("Adding library paths from .gpirc file")
        search_dirs += Config.GPI_LIBRARY_PATH
    elif options.ignore_sys:
        print("Adding self as a library path. No other node libraries will be used")
        search_dirs += [CWD] # 
    else:
        # resort to searching the CWD for libraries
        # -if the make is being invoked on a PyMOD is reasonable to assume there
        # is a library that contains this file potentially 2 levels up.
        print("Looking two levels up from current working directory for node library files")
        search_dirs = [CWD, os.path.realpath(CWD+'/../../')]

    for flib in search_dirs:
        if os.path.isdir(flib): # skip default config if dirs dont exist
            for usrdir in findLibraries(flib):
                p = os.path.dirname(usrdir)
                b = os.path.basename(usrdir)

                if (b in list(found_libs.keys())) and not (p in list(found_libs.values())):
                    print((Cl.FAIL + "ERROR: \'" + str(b) + "\' libraray conflict:"+Cl.ESC))
                    print(("\t "+os.path.join(found_libs[b],b)))
                    print(("\t "+os.path.join(p,b)))
                    sys.exit(ERROR_LIBRARY_CONFLICT)

                msg = "\tGPI_LIBRARY_PATH \'"+str(p)+"\' for lib \'"+str(b)+"\'"
                include_dirs += [os.path.dirname(usrdir)]
                found_libs[b] = p
                print(msg)

    if len(list(found_libs.keys())) == 0:
        print((Cl.WRN + "WARNING: No GPI libraries found!\n" + Cl.ESC))

    if options.preprocess:
        extra_compile_args.append('-E')

    if options.suppressWarnings:
        extra_compile_args.append('-w')

    # debug pyfi arrays
    if options.debug:
        print("Turning on PyFI Array Debug")
        extra_compile_args += ['-DPYFI_ARRAY_DEBUG']

    # fftw_threads is included in the main lib file on Windows
    if platform.system() == 'Windows':
        libraries += ['fftw3', 'fftw3f']
    else:
        libraries += ['fftw3_threads', 'fftw3', 'fftw3f_threads', 'fftw3f']

    # POSIX THREADS
    # this location is the same for Ubuntu and OSX
    print("Adding POSIX-Threads lib")
    if platform.system() == 'Windows':
        libraries += ['pthreads']
    else:
        libraries += ['pthread']
    if not options.ignore_sys:
        print("Adding POSIX-Threads lib")
        include_dirs += ['/usr/include']
        library_dirs += ['/usr/lib']

    # The intel libs and extra compile flags are different between linux and OSX
    if platform.system() == 'Linux':
        pass

    elif platform.system() == 'Darwin':  # OSX

        os.environ["CC"] = 'clang'
        os.environ["CXX"] = 'clang++'

        # force only x86_64
        os.environ["ARCHFLAGS"] = '-arch x86_64'

        # force 10.9 compatibility unless override is passed
        if options.osx_target_ver is not None:
            os.environ["MACOSX_DEPLOYMENT_TARGET"] = options.osx_target_ver
        else:
            os.environ["MACOSX_DEPLOYMENT_TARGET"] = '10.9'

        # for malloc.h
        if not options.ignore_sys:
            include_dirs += ['/usr/include/malloc']

        # default g++
        extra_compile_args += ['-Wsign-compare']

        # unsupported g++
        #extra_compile_args += ['-Wuninitialized']

        # warn about implicit down casting
        #extra_compile_args += ['-Wshorten-64-to-32']

    # COMPILE
    successes = []
    failures = []
    py_successes = []
    py_failures = []
    for target in targets:

        os.chdir(target['pth'])

        # PYTHON regression, error checking, pep8
        if target['ext'] == '.py':
            retcode = makePy(target['fn'], target['ext'],
                             fmt=options.format,
                             check_fmt=options.check_format)

            if retcode != 0:
                py_failures.append(target['fn'])
            else:
                py_successes.append(target['fn'])

        else:  # CPP compilation

            # ASTYLE
            if options.format:
                try:
                    print("\nAstyle...")
                    print("Reformatting CPP Code: " + target['fn'] + target['ext'])
                    # TODO: astyle might not be in the path even if on the system
                    os.system('astyle -A1 -S -w -c -k3 -b -H -U -C '
                              + target['fn'] + target['ext'])
                    continue  # don't proceed to compile
                except:
                    print("Failed to perform auto-formatting with \'astyle\'. Do you have it installed?")
                    sys.exit(ERROR_EXTERNAL_APP)

            if default_cpp:
                extra_compile_args.append('-std=c++11')

            mod_name = target['fn'].split("_PyMOD")[0]
            extra_compile_args.append('-DMOD_NAME=' + mod_name)

            retcode = compile(
                mod_name, include_dirs, libraries, library_dirs,
                extra_compile_args, runtime_library_dirs)

            extra_compile_args.pop()  # remove MOD_NAME for the next target

            if retcode != 0:
                failures.append(target['fn'])
            else:
                successes.append(target['fn'])

    show_summary = len(py_successes) + len(py_failures) + len(successes) + len(failures)

    # Py Summary
    if show_summary > 1:
        print(('\nSUMMARY (Py Compilations):\n\tSUCCESSES ('+Cl.OKGR+str(len(py_successes))+Cl.ESC+'):'))
        for i in py_successes:
            print(("\t\t" + i))
        print(('\tFAILURES ('+Cl.FAIL+str(len(py_failures))+Cl.ESC+'):'))
        for i in py_failures:
            print(("\t\t" + i))

    # CPP Summary
    if show_summary > 1:
        print(('\nSUMMARY (CPP Compilations):\n\tSUCCESSES ('+Cl.OKGR+str(len(successes))+Cl.ESC+'):'))
        for i in successes:
            print(("\t\t" + i))
        print(('\tFAILURES ('+Cl.FAIL+str(len(failures))+Cl.ESC+'):'))
        for i in failures:
            print(("\t\t" + i))

    # ON FAILURE
    if (len(py_failures) + len(failures)) > 0:
        sys.exit(ERROR_FAILED_COMPILATION)
    # ON SUCCESS
    else:
        sys.exit(SUCCESS)

if __name__ == '__main__':
    make()



#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

'''
Classes for loading and saving each version of gpi networks.  Each class
corresponds to a particular network file description.  Upon load, the network
file type is determined and then the appropriate class is installed. On write,
the latest network description is used.
'''


# TODO: add a header to the pickle file
#       -create a gpickle class that pickles the network description
#        and prepends version info before the pickle info


import os
import re
import sys
import json
import time
import codecs
import pickle
import traceback

# gpi
from gpi import QtGui, QtCore, VERSION
from .config import Config
from .defines import GetHumanReadable_time, GetHumanReadable_bytes, TranslateFileURI
from .logger import manager
from .sysspecs import Specs
from .widgets import GPIFileDialog

# start logger for this module
log = manager.getLogger(__name__)

def convert_keysandvals_to_string(dictionary):
    """Recursively converts unicode vals to strings.
    Operates on dictionaries, lists and dict-keys.
    """
    if not isinstance(dictionary, dict):
        if isinstance(dictionary, unicode):
            return str(dictionary.encode('latin1'))
        if isinstance(dictionary, list):
            return list(convert_keysandvals_to_string(k) for k in dictionary)
        return dictionary
    return dict((str(k.encode('latin1')), convert_keysandvals_to_string(v)) for k, v in dictionary.items())

class Network_Base(object):
    '''All networks should implement this base class and be named with the
    'Network_v<version>' where 'version' is a str(integer).
    '''
    GPINET_VERSION=''

    def __init__(self, fname, contents=None):
        # fname: filename for saving or loading
        # contents: an object containing network info compatible with the canvas'
        #   serializeCanvas() and deserializeCanvas() routines.
        pass

    def load(self):
        # open the file and load its contents into memory
        # -via pickle, config, xml, etc...
        pass

    def save(self):
        # open a file for writing and write the network object contents to disk
        pass

    def test(self):
        # test to see if the network file is THIS version, return True if it is.
        return False

    def version(self):
        # return the version number as a str
        return str(self.GPINET_VERSION)

    def __str__(self):
        # for sorting
        return self.version()

class Network_v1(Network_Base):
    '''The first version of the network interface compatible with the released
    GPI-beta.  The file format is pickled list() and dict() objects.  The original 
    released header doesn't necessarily contain a version number so network
    files w/o are assumed to be this version.
    '''
    GPINET_VERSION='1'

    def __init__(self, fname, contents=None):
        self._fname = fname  # not actually needed since parent did the loading
        self._contents = contents  # a shortcut if the caller did the reading
                                   # also used for storing network data to save

    def test(self):
        # there is no way of verifying the first network version so it always
        # returns false and is only invoked as a last resort.
        # the best we can do is read in the whole file and check some fields
        try:
            with open(self._fname, "rb") as fptr:
                contents = pickle.load(fptr, encoding="latin1")
        except:
            log.debug('Network_v1 test: '+str(traceback.format_exc()))
            return False

        # check for dictionary structure
        if not isinstance(contents, dict):
            log.debug('Network_v1 test: Failure: contents are not of type dict.')
            return False
        return True

    def load(self):
        # load file contents into memory
        with open(self._fname, "rb") as fptr:
            self._contents = pickle.load(fptr, encoding="latin1")
        return self.convert_incoming()

    def save(self):
        # save network contents to a file
        # convert to the right type
        self.convert_outgoing()
        try:
            with open(self._fname, "wb") as fptr:
                pickle.dump(self._contents, fptr)
            log.dialog("Network saved.")
        except:
            log.error("Saving network failed.")
            log.error(traceback.format_exc())

    def convert_incoming(self):
        # From file.
        # convert loaded data into common object format
        # -since its pickled its already in the right format

        # take this opportunity to print out some network file stats
        msg = 'Network file info:\n'
        if 'NETWORK_VERSION' in self._contents:
            msg += '\tnet-version: '+str(self._contents['NETWORK_VERSION']) + '\n'
        else:
            msg += '\tassumed net-version: '+str(self.version())+'\n'

        if 'GPI_VERSION' in self._contents:
            msg += '\tsaved with gpi-version: '+str(self._contents['GPI_VERSION']) + '\n'
        log.warn(msg)

        # BACKWARD COMPATIBILITY
        # add a dummy node key into all the node dicts
        # -since the key is an empty string, the lookup will fall back to
        #  finding the best match
        for node in self._contents['nodes']['nodes']:
            node['key'] = ''
            node['walltime'] = '0'
            node['avgwalltime'] = '0'
            node['stdwalltime'] = '0'

        return self._contents

    def convert_outgoing(self):
        # Takes input from canvas.  To file.
        # convert serialized data in memory to the type required by the file
        # format.  Since its a nested dict(), pickle can handle this directly.
        network = self._contents

        # network files are dictionaries that contain:
        # {'nodes': <a list of node settings>}
        #   -widget info
        #   -port info
        #   -connections
        #   -position
        #   -TODO: menu window position
        #   -TODO: original node path (not yet sure if
        #               specificity is better here)
        #
        # {'path': <a list of paths where nodes can be found>}
        #   -sys.path (accumulated dragNdrop, loads, etc...)
        #
        # Potential Future Additions:
        # {'canvas': <a list of canvas params>}
        #   -size, shape (depends on display)
        #network['path'] = sys.path  # no longer needed
        network['NETWORK_VERSION'] = self.version()
        network['GPI_VERSION'] = VERSION
        network['HEADER'] = 'This is a GPI Network File'

class Network_v2(Network_Base):
    '''The second version of the network interface.  This version supports new
    2-level library scope which exists as an extra tag in the node settings dict.
    The file format is pickled list() and dict() objects.  The original 
    released header doesn't necessarily contain a version number so network
    files w/o are assumed to be this version.
    '''
    GPINET_VERSION='2'

    def __init__(self, fname, contents=None):
        self._fname = fname  # not actually needed since parent did the loading
        self._contents = contents  # a shortcut if the caller did the reading
                                   # also used for storing network data to save

    def test(self):
        # the best we can do is read in the whole file and check some fields
        try:
            with open(self._fname, "rb") as fptr:
                contents = pickle.load(fptr, encoding="latin1")
        except:
            log.debug('Network_v2 test: '+str(traceback.format_exc()))
            return False

        # check for dictionary structure
        if not isinstance(contents, dict):
            log.debug('Network_v2 test: Failure: contents are not of type dict.')
            return False

        # check the version
        if 'NETWORK_VERSION' in contents:
            if str(contents['NETWORK_VERSION']) == self.version():
                return True
        return False

    def load(self):
        # load file contents into memory
        with open(self._fname, "rb") as fptr:
            self._contents = pickle.load(fptr, encoding="latin1")
        return self.convert_incoming()

    def save(self):
        # save network contents to a file
        # convert to the right type
        self.convert_outgoing()
        try:
            with open(self._fname, "wb") as fptr:
                pickle.dump(self._contents, fptr)
            log.dialog("Network saved.")
        except:
            log.error("Saving network failed.")
            log.error(traceback.format_exc())

    def convert_incoming(self):
        # From file.
        # convert loaded data into common object format
        # -since its pickled its already in the right format

        # take this opportunity to print out some network file stats
        msg = 'Network file info:\n'
        if 'NETWORK_VERSION' in self._contents:
            msg += '\tnet-version: '+str(self._contents['NETWORK_VERSION']) + '\n'

        if 'GPI_VERSION' in self._contents:
            msg += '\tsaved with gpi-version: '+str(self._contents['GPI_VERSION']) + '\n'

        if 'DATETIME' in self._contents:
            msg += '\tdate saved: '+str(self._contents['DATETIME']) + '\n'

        if 'WALLTIME' in self._contents:
            msg += '\twall time: '+str(GetHumanReadable_time(float(self._contents['WALLTIME']))) + '\n'

        if 'TOTAL_PMEM' in self._contents:
            msg += '\ttotal port mem: '+str(GetHumanReadable_bytes(int(self._contents['TOTAL_PMEM']))) + '\n'

        if 'PLATFORM' in self._contents:
            for k,v in list(self._contents['PLATFORM'].items()):
                msg += '\t'+k+': '+str(v)+'\n'

        log.dialog(msg)

        return self._contents

    def convert_outgoing(self):
        # Takes input from canvas.  To file.
        # convert serialized data in memory to the type required by the file
        # format.  Since its a nested dict(), pickle can handle this directly.
        network = self._contents

        # network files are dictionaries that contain:
        # {'nodes': <a list of node settings>}
        #   -widget info
        #   -port info
        #   -connections
        #   -position
        #   -TODO: menu window position
        #   -TODO: original node path (not yet sure if
        #               specificity is better here)
        #
        # {'path': <a list of paths where nodes can be found>}
        #   -sys.path (accumulated dragNdrop, loads, etc...)
        #
        # Potential Future Additions:
        # {'canvas': <a list of canvas params>}
        #   -size, shape (depends on display)
        #network['path'] = sys.path  # no longer needed
        network['NETWORK_VERSION'] = self.version()
        network['GPI_VERSION'] = VERSION
        network['HEADER'] = 'This is a GPI Network File'
        network['DATETIME'] = str(time.asctime(time.localtime()))

        # network['PLATFORM']
        network['PLATFORM'] = Specs.table()

class Network_v3(Network_v2):
    '''The third version of the network interface.  This version supports
    the json data file format for better multi-platform support.
    '''
    GPINET_VERSION='3'

    def __init__(self, fname, contents=None):
        super(Network_v3, self).__init__(fname, contents)
        self._fname = fname  # not actually needed since parent did the loading
        self._contents = contents  # a shortcut if the caller did the reading
                                   # also used for storing network data to save

        # write a simple header that can verify the network version directly
        # without having to read the whole file.
        self._header = "# This file was written with GPI v"+str(VERSION)
        self._header += " using Network v"+str(self.version())
        self._header += ". Do not edit this line.\n"

        # validate the network version and potentially the GPI version
        self._header_regex = "(GPI|gpi)\s+v([\w.]+).*[Nn]et.*v([\d]+)"

    def test(self):
        try:
            with open(self._fname, "r", encoding='utf8') as fptr:
                # check the first line in the file
                match = re.compile(self._header_regex).search(fptr.readline())
                if match:
                    if match.group(3) == self.version():
                        # if the versions match, thats gold
                        return True
        except:
            log.debug('Network_v3 test: '+str(traceback.format_exc()))
            return False
        return False

    def save(self):
        # save network contents to a file
        # file operations go here

        # convert to the right type
        self.convert_outgoing()

        try:
            with open(self._fname, "w", encoding='utf8') as fptr:
                fptr.write(self._header)
                json.dump(self._contents, fptr, sort_keys=True, indent=1)
            log.dialog("Network saved.")
        except:
            log.error("Saving network failed. "+str(traceback.format_exc()))

    def load(self):
        # load network dict
        with open(self._fname, "r", encoding='utf8') as fptr:
            fptr.readline() # header
            self._contents = json.load(fptr)
        return self.convert_incoming()


########################################################################
# Network IO Manager Class

class Network(object):
    '''Determines the version of the file to be read, then chooses the correct
    loading object for that version.  The loading object must translate the
    file contents to the various object settings that make up the network
    description.  This object should handle the parent calls.
    '''

    def __init__(self, parent):
        self._parent = parent

        # this is a crutch: see determine_version()
        self._unpickled_contents = None

        self._latest_net_version = None
        self._latest_net_class = None

        # start with this path and then follow the user's cwd for this session
        self._current_working_dir = TranslateFileURI(Config.GPI_NET_PATH)

        self.getLatestClass()

    def getLatestClass(self):
        thismodule = sys.modules[__name__]
        netclasses = []
        for cls in dir(thismodule):
            if cls.startswith('Network_v'):
                netclasses.append(cls)
        versions = [x.split('Network_v')[1] for x in netclasses]

        self._latest_net_version = max(versions)
        self._latest_net_class = getattr(thismodule, 'Network_v'+self._latest_net_version)

    def netDesc(self):
        # all network objects that derive from Network_Base()
        l = []
        for mnam in dir(sys.modules[__name__]):
            mod = getattr(sys.modules[__name__], mnam)
            if hasattr(mod, 'GPINET_VERSION'):
                l.append(mod)
        l = sorted(l, key=str, reverse=True) # try highest version first
        return l

    def determine_version(self, fname):
        # check the file format for version and format info
        # -loading is the only way to check if its pickled, so just return
        # the unpickled contents along with the version.  In the future, if
        # pickle is no longer used, then the other format will be checked
        # first, then pickle as a backup.

        # loop over all of the Network description classes
        for obj in self.netDesc():
            # test the network file with the given class
            n = obj(fname)
            try:
                if n.test():
                    # set the version number if the network passed
                    version = n.version()
                    break
            except pickle.UnpicklingError:
                pass

        if version != self._latest_net_version:
            log.warn('This network was saved in an older format, please re-save this network.')

        return version

    def loadNetworkFromFile(self, fname):
        # Used with drops, command-line input, and dialog input
        log.debug('loadNetworkFromFile()')

        # check for valid filename
        if not os.path.isfile(fname):
            log.error("\'" + str(fname) + "\' is not a valid filename, skipping.")
            return

        # determine network format obj type by version and load contents
        # TODO: make this automatically choose the right loader object
        log.debug('determine_version()')
        ver = self.determine_version(fname)
        log.debug('version found: '+str(ver))
        if ver == self._latest_net_version:
            net = self._latest_net_class(fname, contents=self._unpickled_contents)
        elif hasattr(sys.modules[__name__], 'Network_v'+ver):
            net = getattr(sys.modules[__name__], 'Network_v'+ver)(fname, contents=self._unpickled_contents)
        else:
            log.error('Network version cannot be determined, skipping.')
            return

        log.debug('net.load() called')
        return net.load()  # network data objects

    def saveNetworkToFile(self, fname, network):
        # used for dialog, possibly command-line

        # enforce network filename consistency
        if not fname.endswith('.net'):
            fname += '.net'

        # always save in the latest version
        net = self._latest_net_class(fname, contents=network)
        net.save()

    def loadNetworkFromFileDialog(self):

        # start looking in user config'd network dir
        #start_path = os.path.expanduser(self._parent.parent._networkDir)
        #start_path = os.path.expanduser('~/')
        start_path = TranslateFileURI(Config.GPI_NET_PATH)
        log.info("File browser start path: " + start_path)

        # create dialog box
        kwargs = {}
        kwargs['filter'] = 'GPI network (*.net)'
        kwargs['caption'] = 'Open Session (*.net)'
        kwargs['directory'] = self._current_working_dir
        dia = GPIFileDialog(self._parent, **kwargs)

        # don't run if cancelled
        if dia.runOpenFileDialog():

            # save the current directory for next browse
            if Config.GPI_FOLLOW_CWD:
                self._current_working_dir = str(dia.directory().path())

            fname = str(dia.selectedFiles()[0])

            return self.loadNetworkFromFile(fname)

    def listMediaDirs(self):
        if Specs.inOSX():
            rdir = '/Volumes'
            if os.path.isdir(rdir):
                return ['file://'+rdir+'/'+p for p in os.listdir(rdir)]
        elif Specs.inLinux():
            rdir = '/media'
            if os.path.isdir(rdir):
                return ['file://'+rdir+'/'+p for p in os.listdir(rdir)]
            rdir = '/mnt'
            if os.path.isdir(rdir):
                return ['file://'+rdir+'/'+p for p in os.listdir(rdir)]
        return []

    def saveNetworkFromFileDialog(self, network):

        # make sure there is something worth saving
        nodes = network['nodes']
        layouts = network['layouts']

        log.debug(str(nodes))
        log.debug(str(layouts))

        if nodes is None:
            # Error message to user, should go to statusbar
            log.warn("No network data to save, skipping.")
            return

        if not len(nodes['nodes']) and not len(nodes['macroNodes']):
            # Error message to user, should go to statusbar
            log.warn("No network data to save, skipping.")
            return

        kwargs = {}
        kwargs['filter'] = 'GPI network (*.net)'
        kwargs['caption'] = 'Save Session (*.net)'
        kwargs['directory'] = self._current_working_dir
        dia = GPIFileDialog(self._parent, **kwargs)
        dia.selectFile('Untitled.net')

        # don't run if cancelled
        if dia.runSaveFileDialog():

            # save the current directory for next browse
            if Config.GPI_FOLLOW_CWD:
                self._current_working_dir = str(dia.directory().path())

            fname = dia.selectedFilteredFiles()[0]
            self.saveNetworkToFile(fname, network)



#!/opt/gpi/bin/python

#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

"""A simple test script for the PyFI eigen wrapper.
"""
import numpy as np
import eigen
import timeit

if __name__ == "__main__":
    m = 500
    n = 500
    A = 1000*np.random.random((m,n))
    npAinv = np.linalg.pinv(A)
    eigenAinv = eigen.pinv(A)
    if np.allclose(npAinv, eigenAinv, rtol=0.01):
        init = """import eigen, numpy as np
m,n = 50,50
A = 1000*np.random.random((m,n))"""
        a = timeit.timeit('np.linalg.pinv(A)', setup=init, number=1000)
        b = timeit.timeit('eigen.pinv(A)', setup=init, number=1000)
        print("eigen successfully inverted the {}x{} matrix {} times faster than numpy".format(m,n,a/b))

    p = 3
    B = 1000*np.random.random((m,p))
    eigenX = eigen.solve(A,B)
    npX = np.linalg.lstsq(A,B)[0]
    if np.allclose(eigenX, npX, rtol=0.01):
        init = """import eigen as eigen, numpy as np
m,n,p = 50,50,3
A = 1000*np.random.random((m,n))
B = 1000*np.random.random((m,p))"""
        a = timeit.timeit('np.linalg.lstsq(A,B)', setup=init, number=1000)
        b = timeit.timeit('eigen.solve(A,B)', setup=init, number=1000)
        print("eigen successfully solved the equatioon AX=B {} times faster than numpy\n\twhere A was {}x{}, and B was {}x{}".format(a/b,m,n,m,p))


    p = 2000
    B = 1000*np.random.random((n,p))
    eigenC = eigen.dot(A,B)
    npC = np.dot(A,B)
    if np.allclose(eigenC, npC, rtol=0.01):
        init = """import eigen as eigen, numpy as np
m,n,p = 50,50,20
A = 1000*np.random.random((m,n))
B = 1000*np.random.random((n,p))"""
    a = timeit.timeit('np.dot(A,B)', setup=init, number=1000)
    b = timeit.timeit('eigen.dot(A,B)', setup=init, number=1000)
    print("eigen successfully multiplied A*B {} times faster than numpy\n\twhere A was {}x{}, and B was {}x{}".format(a/b,m,n,n,p))




#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

# Brief: Commandline option parsing.
#           -can initiate some simple commands

import sys
import logging
import optparse

# gpi
from .associate import isGPIAssociatedFile
from .defines import isGPIModFile, isGPINetworkFile
from gpi import VERSION
from . import logger
from .logger import manager
from .sysspecs import Specs

# start logger for this module
log = manager.getLogger(__name__)


class CmdParser(object):
    '''An object to parse input commandline args after the QApplication has
    done its own parsing.
    '''

    def __init__(self):
        '''Initialize all command-line parsed variables.
        '''
        self._argv = None
        self._options = None
        self._args = None

        # a place for string args option
        self._sargs = {}

        # no gui option
        self._nogui = False
        self._scriptMode = False

        # splash is on by default
        self._nosplash = False

        self._loadable_mods = []
        self._loadable_nets = []
        self._loadable_files = []  # associated files

        # tell the USAGE text that this is always 'gpi'
        self._parser = optparse.OptionParser(prog='gpi')

        # take in any filename for extension checking, then loading. 
        self._parser.add_option('--config', dest='dumpConfig', action='store_true', help='''GPI will read the User ENV and config file and dump the parsed info to stdout.''')
        self._parser.add_option('--log', dest='loglevel', action='store', choices=['debug', 'info', 'node', 'warn', 'error', 'critical'], help='''Change the output level of the logger: debug, info, node, warn, error, and critical''')
        self._parser.add_option('--nogui', dest='nogui', action='store_true', help='''causes GPI to run without a GUI for scripting.  Requires a network file.  The --script option is implied.''')
        self._parser.add_option('--script', dest='script', action='store_true', help='''causes GPI to terminate after the supplied network is finished executing.  Requires a network file.''')
        self._parser.add_option('-s', '--string', dest='string', action='append', type='string', default=[], help='''passes a string arg to a String-node by label.  Handles multiple args.  Syntax: -s <label1>:<string/path> -s <label2>:<string/path>.''')
        self._parser.add_option('--specs', dest='dumpSpecs', action='store_true', help='''GPI will create a platform specs file and exit.''')
        self._parser.add_option('--defines', dest='dumpDefines', action='store_true', help='''Show some internally used defines, such as temp directory paths.''')
        self._parser.add_option('--nosplash', dest='nosplash', action='store_true', help='''Skip the splash screen.''')

    def parse(self, argv):
        # keep a copy of what was parsed
        self._argv = list(argv)

        # options: processed args that use switches
        # args: leftover positional args
        self._options, self._args = self._parser.parse_args(self._argv)

        # check for loadable files
        self.checkArgsForFiles()

        # check and process string args
        self.storeStringNodeArgs()

        # make sure the user passes a network
        if self._options.nogui:
            if self.netCount():
                self._nogui = True
            else:
                log.error('the --nogui option was passed without a network, exiting.')
                sys.exit(1)

        # make sure the user passes a network
        if self._options.script:
            if self.netCount():
                self._scriptMode = True
            else:
                log.error('the --script option was passed without a network, exiting.')
                sys.exit(1)

        # set log level asap, don't wait for mainWindow to be set
        if self.logLevel():
            logger.manager.setLevel(self.logLevel())

        # run simple commands
        if self._options.dumpSpecs:
            self.dumpSpecs()

        if self._options.dumpDefines:
            self.dumpDefines()

        if self._options.dumpConfig:
            from .config import Config
            log.dialog('Config:\n'+str(Config))
            sys.exit(0)

        # splash
        self._nosplash = self._options.nosplash

    def dumpDefines(self):
        import gpi.defines
        msg = []
        for d in dir(gpi.defines):
            o = getattr(gpi.defines,d)
            if type(o) is str:
                if d.startswith('GPI'):
                    msg.append(d + ': ' + o + '\n')
        msg = sorted(msg)
        msg = '\n'+''.join(msg)
        log.dialog(msg)
        sys.exit(0)

    def dumpSpecs(self):
        with open('specs.txt', 'wb') as specsfile:
            specsfile.write('# GPI (v'+str(VERSION)+') system specifications file.\n')
            for k,v in list(Specs.table().items()):
                msg = k+': '+str(v) + '\n'
                specsfile.write(msg)

        log.dialog('Specs file written, exiting.')
        sys.exit(0)

    def logLevel(self):
        if self._options.loglevel == 'debug':
            return logging.DEBUG
        if self._options.loglevel == 'info':
            return logging.INFO
        if self._options.loglevel == 'node':
            return logger.GPINODE
        if self._options.loglevel == 'warn':
            return logging.WARNING
        if self._options.loglevel == 'error':
            return logging.ERROR
        if self._options.loglevel == 'critical':
            return logging.CRITICAL

    def noGUI(self):
        return self._nogui

    def scriptMode(self):
        return self._scriptMode

    def noSplash(self):
        return self._nosplash

    def mods(self):
        return self._loadable_mods

    def nets(self):
        return self._loadable_nets

    def files(self):
        return self._loadable_files

    def stringNodeArg(self, key):
        if key in self._sargs:
            return self._sargs[key]
        log.error('\''+str(key)+'\' not found in string args.')

    def stringNodeLabels(self):
        return list(self._sargs.keys())

    def storeStringNodeArgs(self):
        # merge any redundant labels with warnings
        for arg in self._options.string:
            lab, path = arg.split(':')
            if lab in self._sargs:
                log.warn('input string label for arg: \''+ str(arg) + '\' already exists, skipping.') 
            else:
                self._sargs[lab] = path

    def __str__(self):
        msg = ''
        msg += 'Unparsed: '+str(self._args) + '\n'
        msg += 'Parsed: '+str(self._options) + '\n'
        return msg

    def modCount(self):
        return len(self._loadable_mods)

    def netCount(self):
        return len(self._loadable_nets)

    def fileCount(self):
        return len(self._loadable_files)

    def stringNodeArgCount(self):
        return len(self._sargs)

    def pendingCount(self):
        return self.modCount() + self.netCount() + self.fileCount() + self.stringNodeArgCount()

    def checkArgsForFiles(self):

        for arg in self._args:

            if isGPIModFile(arg):
                self._loadable_mods.append(arg)

            elif isGPINetworkFile(arg):
                self._loadable_nets.append(arg)

            elif isGPIAssociatedFile(arg):
                self._loadable_files.append(arg)

Commands = CmdParser()



#    Copyright (C) 2021 Mayo Clinic
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

from __future__ import print_function

import time
import grpc
import dill
from . import gpi_pb2
from . import gpi_pb2_grpc
from .logger import manager
dill.settings['recurse'] = True

# start logger for this module
log = manager.getLogger(__name__)

options = [
        ('grpc.max_send_message_length', -1),
        ('grpc.max_receive_message_length', -1),
        ("grpc.so_reuseport", 1),
        ("grpc.use_local_subchannel_pool", 1),
    ]

def run(f, args):
    channel = grpc.insecure_channel('archimedes.mayo.edu:50051', options=options)
    stub = gpi_pb2_grpc.GPIStub(channel)
    func = dill.dumps(f)
    args = dill.dumps(args)
    start = time.time()
    response = stub.run(gpi_pb2.Request(function=func, args=args))
    end = time.time()
    print(end- start)
    return dill.loads(response.result)

def run_on_server(function=None, parallel=False):
    from gpi import parallel as par
    def decorator(func):
        def wrapper(*args, **kwargs):
            if parallel:
                result = par.function(run)(func, args)
            else:
                result = run(func, args)
            return result
        wrapper.func = func
        return wrapper
    if function: return decorator(function)
    return decorator